"""Conformer tools."""

from copy import copy, deepcopy
import functools
import numbers
from os import PathLike
from pathlib import Path
import subprocess
import tempfile
import typing
from typing import (
    Any,
    cast,
    Counter,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
import warnings

import numpy as np
from pkg_resources import parse_version

from morfeus.data import (
    HARTREE,
    HARTREE_TO_KCAL,
    HARTREE_TO_KJ,
    K_B,
    KCAL_TO_HARTREE,
    KJ_TO_HARTREE,
)
from morfeus.io import CrestParser, write_xyz
from morfeus.qc import optimize_qc_engine, sp_qc_engine
from morfeus.typing import (
    Array1D,
    Array2D,
    Array3D,
    ArrayLike1D,
    ArrayLike2D,
    ArrayLike3D,
)
from morfeus.utils import (
    convert_elements,
    Import,
    requires_dependency,
    requires_executable,
)

if typing.TYPE_CHECKING:
    import openbabel
    import openbabel.openbabel as ob
    import openbabel.pybel as pybel
    import rdkit
    from rdkit import Chem
    from rdkit.Chem import AllChem, Descriptors
    import spyrmsd


def boltzmann_average_dT(
    properties: ArrayLike1D, energies: ArrayLike1D, temperature: float = 298.15
) -> float:
    """Return the derivative of the Boltzmann average.

    Args:
        properties: Conformer properties
        energies: Conformer energies (a.u.)
        temperature: Temperature (K)

    Returns:
        derivative: Derivative of Boltzmann average.
    """
    energies = np.array(energies)

    # Calculate Boltzmann averaged properties
    avg_prop_en = boltzmann_statistic(
        properties * energies * HARTREE,
        energies,
        temperature=temperature,
        statistic="avg",
    )
    avg_en = boltzmann_statistic(
        energies * HARTREE, energies, temperature=temperature, statistic="avg"
    )
    avg_prop = boltzmann_statistic(
        properties, energies, temperature=temperature, statistic="avg"
    )

    # Calculate derivative
    derivative = (avg_prop_en - avg_en * avg_prop) / (K_B * temperature ** 2)

    return derivative


def boltzmann_statistic(
    properties: ArrayLike1D,
    energies: ArrayLike1D,
    temperature: float = 298.15,
    statistic: str = "avg",
) -> float:
    """Compute Boltzmann statistic.

    Args:
        properties: Conformer properties
        energies: Conformer energies (a.u.)
        temperature: Temperature (K)
        statistic: Statistic to compute: 'avg', 'var' or 'std'

    Returns:
        result: Boltzmann statistic
    """
    properties = np.array(properties)

    # Get conformer weights
    weights = boltzmann_weights(energies, temperature)

    # Compute Boltzmann weighted statistic
    result: float
    if statistic == "avg":
        result = np.average(properties, weights=weights)
    elif statistic == "var":
        avg = np.average(properties, weights=weights)
        result = np.sum(weights * (properties - avg) ** 2)
    elif statistic == "std":
        avg = np.average(properties, weights=weights)
        var = np.sum(weights * (properties - avg) ** 2)
        result = np.sqrt(var)

    return result


def boltzmann_weights(energies: ArrayLike1D, temperature: float = 298.15) -> Array1D:
    """Compute Boltzmann weights.

    Args:
        energies: Conformer energies (a.u.)
        temperature: Temperature (K)

    Returns:
        weights: Conformer weights (normalized to unity)
    """
    energies = np.array(energies)
    energies -= energies.min()
    terms = np.exp(-energies / (K_B / HARTREE * temperature))
    weights: np.ndarray = terms / np.sum(terms)

    return weights


class Conformer:
    """Conformer with coordinates, energy and properties.

    Args:
        elemennts: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        energy: Energy (a.u.)
        degeneracy: Degeneracy
        properties: Conformers properties.
        cip_label: Tuple of CIP labels for all atoms: 'R', 'S' or ''

    Attributes:
        cip_label: Tuple of CIP labels for all atoms: 'R', 'S' or ''
        coordinates: Coordinates (Å)
        degeneracy: Degeneracy
        elemennts: Elements as atomic symbols or numbers
        energy: Energy (a.u.)
        properties: Conformers properties.
    """

    cip_label: Optional[Tuple[str, ...]]
    coordinates: Array2D
    degeneracy: int
    elements: Array1D
    energy: Optional[float]
    properties: Dict[str, float]

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        energy: Optional[float] = None,
        degeneracy: int = 1,
        properties: Optional[Dict[str, float]] = None,
        cip_label: Optional[Tuple[str]] = None,
    ) -> None:
        if properties is None:
            properties = {}
        self.coordinates = np.array(coordinates)
        self.elements = np.array(elements)
        self.energy = energy
        self.properties = properties
        self.degeneracy = degeneracy
        self.cip_label = cip_label

    def __repr__(self) -> str:
        n_atoms = self.coordinates.shape[0]
        return f"{self.__class__.__name__}({n_atoms!r} atoms)"


T = TypeVar("T", bound="ConformerEnsemble")


class ConformerEnsemble:
    """Conformer ensemble object.

    Supports sorting, pruning, optimization and single-point calculations.

    Args:
        elements: Elements as atomic symbols or numbers
        conformer_coordinates: Conformer coordinates (Å)
        energies: Energy (a.u.)
        connectivity_matrix: Connectivity matrix
        degeneracies: Degeneracies
        properties: Conformers properties.
        charge: Molecular charge
        multiplicity: Molecular multiplicity.
        formal_charges: Atomic formal charges
        ref_cip_label: Tuple of CIP labels for all atoms: 'R', 'S' or ''

    Attributes:
        charge: Molecular charge
        conformers: Conformers
        connectivity_matrix: Connectivity matrix
        elements: Elements as atomic symbols or numbers
        formal_charges: Atomic formal charges.
        mol: RDKit mol object.
        multiplicity: Molecular multiplicity.
        ref_cip_label: Tuple of CIP labels for all atoms: 'R', 'S' or ''
    """

    charge: int
    confomers: List[Conformer]
    connectivity_matrix: Array2D
    elements: Array1D
    formal_charges: Array1D
    mol: "Chem.Mol"
    multiplicity: int
    ref_cip_label: Optional[Tuple[str, ...]]

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        conformer_coordinates: Optional[ArrayLike3D] = None,
        energies: Optional[ArrayLike1D] = None,
        connectivity_matrix: Optional[ArrayLike2D] = None,
        degeneracies: Optional[ArrayLike1D] = None,
        properties: Optional[Mapping[str, ArrayLike1D]] = None,
        charge: Optional[int] = None,
        multiplicity: Optional[int] = None,
        formal_charges: Optional[ArrayLike1D] = None,
        ref_cip_label: Optional[Tuple[str, ...]] = None,
    ) -> None:
        elements = np.array(convert_elements(elements, output="numbers"))
        self.elements = elements

        if conformer_coordinates is None:
            conformer_coordinates = []
        if formal_charges is None:
            formal_charges = np.zeros(len(elements), dtype=int)
        if charge is None:
            charge = int(np.sum(formal_charges))
        else:
            if charge != np.sum(formal_charges):
                msg = (
                    f"Charge ({charge}) is different from sum of partial "
                    f"charges ({np.sum(formal_charges)})"
                )
                raise Exception(msg)
        if multiplicity is None:
            multiplicity = 1

        # Store conformers
        self.conformers: List[Conformer] = []
        self._add_conformers(conformer_coordinates, energies, degeneracies, properties)

        # Set up attributes
        self.charge = charge
        self.formal_charges = np.array(formal_charges)
        self.multiplicity = multiplicity
        self.connectivity_matrix = np.array(connectivity_matrix)
        self.mol = None
        self.ref_cip_label = ref_cip_label

    def add_conformers(
        self,
        conformer_coordinates: ArrayLike3D,
        energies: Optional[ArrayLike1D] = None,
        degeneracies: Optional[ArrayLike1D] = None,
        properties: Optional[Mapping[str, ArrayLike1D]] = None,
    ) -> "ConformerEnsemble":
        """Add conformer to ensemble.

        Args:
            conformer_coordinates: Conformer coordinates (Å)
            energies: Energies (a.u.)
            degeneracies: Degeneracies
            properties: Conformer properties.

        Returns:
            self: Self
        """
        self._add_conformers(conformer_coordinates, energies, degeneracies, properties)

        return self

    def add_inverted(self) -> "ConformerEnsemble":
        """Add inverted images of all conformers.

        Scrambles stereochemistry and leads to redundant conformers so use with
        care and prune based on RMSD as postprocessing.

        Returns:
            self: Self
        """
        conformers = self.conformers
        conformer_coordinates: List[List[float]] = []
        energies: List[Optional[float]] = []
        for conformer in conformers:
            coordinates = conformer.coordinates * np.array([-1, -1, -1])
            conformer_coordinates.append(coordinates)
            energies.append(conformer.energy)

        if not all([isinstance(energy, float) for energy in energies]):
            self._add_conformers(conformer_coordinates)
        else:
            energies_ = cast(List[float], energies)
            self.add_conformers(conformer_coordinates, energies_)

        return self

    @requires_dependency([Import(module="rdkit.Chem", item="AllChem")], globals())
    def align_conformers(self) -> "ConformerEnsemble":
        """Align conformers with RDKit."""
        self.update_mol()
        AllChem.AlignMolConformers(self.mol)

        return self

    def boltzmann_average_dT(
        self, property_name: str, temperature: float = 298.15
    ) -> float:
        """Calculate temperature derivative of Boltzmann average of property.

        Args:
            property_name: Name of property
            temperature: Temperature (K)

        Returns:
            derivative: Derivative of Boltzmann average
        """
        energies = self.get_energies()
        properties = self.get_properties()[property_name]
        derivative = boltzmann_average_dT(properties, energies, temperature=temperature)

        return derivative

    def boltzmann_statistic(
        self, property_name: str, temperature: float = 298.15, statistic: str = "avg"
    ) -> float:
        """Calculate Boltzmann staistic of property over ensemble.

        Args:
            property_name: Name of property
            statistic: Boltzmann statistic: 'avg', 'var' or 'std'
            temperature: Temperature (K)

        Returns:
            statistic: Boltzmann statistic
        """
        energies = self.get_energies()
        properties = self.get_properties()[property_name]
        statistic = boltzmann_statistic(
            properties, energies, temperature=temperature, statistic=statistic
        )

        return statistic

    def boltzmann_weights(self, temperature: float = 298.15) -> Array1D:
        """Calculate Boltzmann weights for ensemble.

        Args:
            temperature: Temperature (K)

        Returns:
            weights: Conformer weights (normalized to unity)
        """
        energies = self.get_energies()
        weights = boltzmann_weights(energies, temperature)

        return weights

    def detect_enantiomers(
        self, thres: float = 0.01, method: str = "rdkit", include_hs: bool = False
    ) -> Dict[int, List[int]]:
        """Detect enantiomers in ensemble.

        Args:
            thres: RMSD threshold for detecting enantiomers in terms of coordinates
            method: RMSD calculation method: 'obrms-batch', 'obrms-iter', openbabel' or
                'spyrmsd'
            include_hs: Whether to include H atoms when determining enantiomers

        Returns:
            enantiomers: Mapping of enantiomer with conformer id as keys and enantiomer
                ids as values.
        """
        # Add set of inverted conformers
        n_conformers = self.n_conformers
        self.add_inverted()

        # Map conformers to enantiomers
        enantiomer_map: Dict[int, Set[int]] = {i: set() for i in range(n_conformers)}
        rmsds = self.get_rmsd(method=method, include_hs=include_hs, symmetry=False)
        for i in range(n_conformers):
            # Do test that inverted conformer should have 0 RMSD to be
            # enantiomer
            test = np.where(rmsds[i, n_conformers:] < thres)[0]
            if len(test) > 0:
                if i not in test:
                    enantiomer_map[i].update(test)
                    for j in test:
                        enantiomer_map[j].add(i)
        enantiomers = {key: list(value) for key, value in enantiomer_map.items()}

        # Reset number of conformers
        self.conformers = self.conformers[:n_conformers]

        return enantiomers

    @classmethod
    def from_crest(cls: Type[T], path: Union[str, PathLike]) -> T:
        """Generate conformer ensemble from CREST output.

        Args:
            path: Path to CREST folder

        Returns:
            ce: Conformer ensemble object.
        """
        cp = CrestParser(path)
        ce = cls(
            cp.elements,
            cp.conformer_coordinates,
            energies=cp.energies,
            degeneracies=cp.degeneracies,
        )

        return ce

    @classmethod
    def from_rdkit(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        """Generate conformer ensemble from RDKit.

        See the documentation for the function conformers_from_rdkit for
        more information.

        Args:
            *args: Positional arguments
            **kwargs: Keyword arguments

        Returns:
            ce: Conformer ensemble object.
        """
        # Run RDKit conformer search and generate ensemble.
        (
            elements,
            conformer_coordinates,
            energies,
            connectivity_matrix,
            charges,
            mol,
        ) = conformers_from_rdkit(*args, **kwargs)
        ce = cls(
            elements,
            conformer_coordinates,
            energies=energies,
            connectivity_matrix=connectivity_matrix,
            formal_charges=charges,
        )
        ce.mol = mol
        ce.set_multiplicity_from_mol()

        # Set reference CIP label if enantiomerically pure.
        cip_labels = ce.get_cip_labels()
        if len(set(cip_labels)) == 1:
            ce.ref_cip_label = cip_labels[0]

        return ce

    @classmethod
    def from_ob_ga(
        cls: Type[T],
        *args: Any,
        generate_rdkit_mol: bool = False,
        update_charges: bool = True,
        update_connectivity: bool = True,
        update_multiplicity: bool = True,
        **kwargs: Any,
    ) -> T:
        """Generate conformer ensemble from OpenBabel with GA method.

        See the documentation for the function conformers_from_ob_ga
        for more information.

        Args:
            *args: Positional arguments for conformers_from_ob_ga
            generate_rdkit_mol: Generate RDKit mol object for ensemble
            update_charges: Update formal charges from generated RDKit Mol object. Only
                used if generate_rdkit_mol is True.
            update_connectivity: Update connectivity from generated RDKit Mol object.
                Only used if generate_rdkit_mol is True.
            update_multiplicity: Update multiplicity from generated RDKit Mol object.
                Only used if generate_rdkit_mol is True.
            **kwargs: Keyword arguments for conformers_from_ob_ga

        Returns:
            ce: Conformer ensemble.
        """
        # Run Openbabel conformer search and generate ensemble.
        (
            elements,
            conformer_coordinates,
            connectivity_matrix,
            charges,
            ob_mol,
        ) = conformers_from_ob_ga(*args, **kwargs)
        ce = cls(
            elements,
            conformer_coordinates,
            connectivity_matrix=connectivity_matrix,
            formal_charges=charges,
        )

        # Generate RDKit mol object and CIP labels
        if generate_rdkit_mol:
            ce.generate_mol(
                update_charges=update_charges,
                update_connectivity=update_connectivity,
                update_multiplicity=update_multiplicity,
            )

            cip_labels = ce.get_cip_labels()
            if len(set(cip_labels)) == 1:
                ce.ref_cip_label = cip_labels[0]

        return ce

    @classmethod
    def from_ob_ff(
        cls: Type[T],
        *args: Any,
        generate_rdkit_mol: bool = False,
        update_charges: bool = True,
        update_connectivity: bool = True,
        update_multiplicity: bool = True,
        **kwargs: Any,
    ) -> T:
        """Generate conformer ensemble from OpenBabel with FF method.

        See the documentation for the function conformers_from_ob_ff
        for more information.

        Args:
            *args: Positional arguments for conformers_from_ob_ga
            generate_rdkit_mol: Generate RDKit mol object for ensemble
            update_charges: Update formal charges from generated RDKit Mol object. Only
                used if generate_rdkit_mol is True.
            update_connectivity: Update connectivity from generated RDKit Mol object.
                Only used if generate_rdkit_mol is True.
            update_multiplicity: Update multiplicity from generated RDKit Mol object.
                Only used if generate_rdkit_mol is True.
            **kwargs: Keyword arguments for conformers_from_ob_ga

        Returns:
            ce: Conformer ensemble.
        """
        # Run Openbabel conformer search and generate ensemble.
        (
            elements,
            conformer_coordinates,
            connectivity_matrix,
            charges,
            ob_mol,
        ) = conformers_from_ob_ff(*args, **kwargs)
        ce = cls(
            elements,
            conformer_coordinates,
            connectivity_matrix=connectivity_matrix,
            formal_charges=charges,
        )

        # Generate RDKit mol object and CIP labels
        if generate_rdkit_mol:
            ce.generate_mol(
                update_charges=update_charges,
                update_connectivity=update_connectivity,
                update_multiplicity=update_multiplicity,
            )

            cip_labels = ce.get_cip_labels()
            if len(set(cip_labels)) == 1:
                ce.ref_cip_label = cip_labels[0]

        return ce

    @requires_dependency([Import(module="rdkit", item="Chem")], globals())
    def generate_mol(
        self,
        update_charges: bool = True,
        update_connectivity: bool = True,
        update_multiplicity: bool = True,
    ) -> None:
        """Generate RDKit Mol object.

        Args:
            update_charges: Update formal charges from generated RDKit Mol object
            update_connectivity: Update connectivity from generated RDKit Mol object
            update_multiplicity: Update multiplicity from generated RDKit Mol object
        """
        mol = _get_rdkit_mol(
            self.elements,
            self.get_coordinates(),
            self.connectivity_matrix,
            self.formal_charges,
        )
        if update_charges:
            self.formal_charges = np.array(
                [atom.GetFormalCharge() for atom in mol.GetAtoms()]
            )
        if update_connectivity:
            self.connectivity_matrix = np.array(
                Chem.GetAdjacencyMatrix(mol, useBO=True)
            )
        if update_multiplicity:
            self.set_multiplicity_from_mol()

        self.mol = mol

    @requires_dependency([Import(module="rdkit", item="Chem")], globals())
    def get_cip_labels(self) -> List[Tuple[str, ...]]:
        """Generate tuples of CIP labels for conformer.

        Returns:
            cip_labels: Tuples of CIP labels for each conformer.
        """
        # Update RDKit Mol object with current conformers.
        self.update_mol()

        # Generate CIP labels with new RDKit
        cip_labels = []
        mol = self.mol
        for i in range(mol.GetNumConformers()):
            Chem.AssignStereochemistryFrom3D(mol, i)
            labels = []
            for atom in mol.GetAtoms():
                if atom.HasProp("_CIPCode"):
                    label = atom.GetProp("_CIPCode")
                else:
                    label = ""
                labels.append(label)
            cip_labels.append(tuple(labels))

        return cip_labels

    def get_coordinates(self) -> Array1D:
        """Get conformers coordinates.

        Returns:
            conformer_coordinates: Conformer coordinates (Å)
        """
        conformer_coordinates = np.array(
            [conformer.coordinates for conformer in self.conformers]
        )

        return conformer_coordinates

    def get_degeneracies(self) -> Array1D:
        """Get conformer degeneracies.

        Returns:
            degeneriacies: Degeneracies
        """
        degeneracies = np.array([conformer.degeneracy for conformer in self.conformers])

        return degeneracies

    def get_energies(self) -> Array1D:
        """Get conformer energies.

        Returns:
            energies: Energy (a.u.)
        """
        energies = np.array([conformer.energy for conformer in self.conformers])
        return energies

    def get_properties(self) -> Dict[str, Array1D]:
        """Get conformer properties.

        Returns:
            properties: Conformer properties
        """
        properties: Dict[str, List[float]] = {}
        for conformer in self.conformers:
            for key, value in conformer.properties.items():
                properties.setdefault(key, []).append(value)
        properties: Dict[str, np.ndarray] = {
            key: np.array(value) for key, value in properties.items()
        }

        return properties

    def get_relative_energies(
        self, unit: str = "kcal/mol", relative: bool = True
    ) -> Array1D:
        """Get conformer energies with choice of units and reference value.

        Args:
            unit: Unit of returned energies: 'hartree', 'kcal/mol' or 'kJ/mol'.
            relative: Return energies relative to lowest-energy conformer.

        Returns:
            energies: Conformer energies.
        """
        energies = self.get_energies()

        # Adjust unit
        if unit.lower() == "hartree":
            pass
        if unit.lower() == "kcal/mol":
            energies *= HARTREE_TO_KCAL
        elif unit.lower() == "kj/mol":
            energies *= HARTREE_TO_KJ

        if relative:
            energies -= np.min(energies)

        return energies

    def get_rmsd(
        self,
        i_s: Optional[Sequence[int]] = None,
        j_s: Optional[Sequence[int]] = None,
        include_hs: bool = False,
        symmetry: bool = False,
        method: str = "rdkit",
    ) -> Array1D:
        """Get RSMD between conformers.

        For very small systems 'openbabel' or 'spyrmsd' work well. For larger systems a
        significant speed-up is attained with 'rdkit', 'obrms-batch' or 'obrms-iter'.

        Args:
            i_s: Indices of conformers
            j_s: Indices of conformers
            include_hs: Whether to include H atoms in RMSD calculation. Ignored for
                'obrms-iter' and 'obrms-batch' that only use heavy atoms.
            symmetry: Consider symmetry (requires connectivity matrix). Ignored for
                'obrms-iter' and 'obrms-batch' that always use symmetry.
            method: RMSD calculation method: 'obrms-batch', 'obrms-iter', 'openbabel',
                'rdkit' or 'spyrmsd'.

        Returns:
            rmsds: RSMDs (Å)
        """
        if i_s is None:
            i_s_ = np.arange(1, len(self.conformers) + 1)
        else:
            i_s_ = np.array(i_s)
        i_s = i_s_
        if j_s is None:
            j_s_ = np.arange(1, len(self.conformers) + 1)
        else:
            j_s_ = np.array(j_s)
        j_s = j_s_

        rmsds: np.ndarray
        if method == "obrms-batch":
            rmsds = self._get_rmsd_obrms_batch(i_s, j_s)
        elif method == "obrms-iter":
            rmsds = self._get_rmsd_obrms_iter(i_s, j_s)
        elif method == "openbabel":
            rmsds = self._get_rmsd_openbabel(i_s, j_s, include_hs, symmetry)
        elif method == "rdkit":
            rmsds = self._get_rmsd_rdkit(i_s, j_s, include_hs)
        elif method == "spyrmsd":
            rmsds = self._get_rmsd_spyrmsd(i_s, j_s, include_hs, symmetry)
        return rmsds

    @property
    def n_conformers(self) -> int:
        """Number of conformers."""
        n_conformers = len(self.conformers)

        return n_conformers

    def optimize_qc_engine(
        self,
        ids: Optional[Sequence[int]] = None,
        program: Optional[str] = None,
        model: Optional[Dict[str, Any]] = None,
        keywords: Optional[Dict[str, Any]] = None,
        local_options: Optional[Dict[str, Any]] = None,
        procedure: str = "berny",
    ) -> "ConformerEnsemble":
        """Optimize conformers with QCEngine interface.

        Args:
            ids: Conformer indices (1-indexed). If None, all are optimized
            program: QCEngine program
            model: QCEngine model
            keywords: QCEngine keywords
            local_options: QCEngine local options
            procedure: QCEngine procedure

        Returns:
            self: Self

        Raises:
            Exception: When RDKit requested without formal charges.
        """
        if program is not None:
            if np.any(self.formal_charges != 0) and (program.lower() == "rdkit"):
                raise Exception("QCEngine using RDKit does not work with charges.")

        # Set defaults
        if model is None:
            model = {}
        if keywords is None:
            keywords = {}
        if local_options is None:
            local_options = {}
        if ids is not None:
            conformers = [self.conformers[i] for i in ids]
        else:
            conformers = self.conformers

        for conformer in conformers:
            opt_coordinates, energies = optimize_qc_engine(
                self.elements,
                conformer.coordinates,
                charge=self.charge,
                multiplicity=self.multiplicity,
                connectivity_matrix=self.connectivity_matrix,
                program=program,
                model=model,
                keywords=keywords,
                local_options=local_options,
                procedure=procedure,
            )
            conformer.coordinates = opt_coordinates
            conformer.energy = energies[-1]

        return self

    def condense_enantiomeric(
        self, thres: Optional[float] = None
    ) -> "ConformerEnsemble":
        """Condense enantiomers into single enantiomer per pair.

        Args:
            thres: RMSD threshold for assessing enantiomers.

        Returns:
            self: Self

        Raises:
            Exception: If molecule is chiral.
        """
        cip_labels = self.get_cip_labels()
        if "".join(["".join(labels) for labels in cip_labels]) != "":
            raise Exception("Chiral molecule. Not safe to condense.")

        enantiomers = self.detect_enantiomers(thres=0.01)
        to_keep = []
        to_remove = []
        for key, value in enantiomers.items():
            if key not in to_remove:
                conformer = self.conformers[key]
                conformer.degeneracy = len(value) + 1
                to_keep.append(conformer)
                to_remove.extend(value)

        self.conformers = to_keep

        return self

    def prune_enantiomers(
        self, keep: str = "original", ref_label: Optional[Tuple[str, ...]] = None
    ) -> "ConformerEnsemble":
        """Prune conformers so that only one enantiomer is present in the ensemble.

        Args:
            keep: Which enantiomer to keep: 'original', 'most common' or 'specified'.
                Choice of 'original' requires that the ref_cip_label attribute is set.
                Choice of 'specified' requires ref_label to be given.
            ref_label: Reference CIP labels for all atoms

        Returns:
            self: Self
        """
        cip_labels = self.get_cip_labels()

        # Set up reference label
        if keep == "original":
            ref_label = self.ref_cip_label
            if ref_label is None:
                keep = "most common"
        elif keep == "specified":
            pass
        if keep == "most common":
            counter: Counter[Tuple[str, ...]] = Counter(cip_labels)
            ref_label = counter.most_common(n=1)[0][0]

        # Prune conformers
        to_keep = []
        for i, cip_label in enumerate(cip_labels):
            if cip_label == ref_label:
                to_keep.append(i)
        self.conformers = [self.conformers[i] for i in to_keep]

        return self

    def prune_energy(
        self, threshold: float = 3.0, unit: str = "kcal/mol"
    ) -> "ConformerEnsemble":
        """Prune conformers based on energy compared to minimum energy conformer.

        Args:
            threshold: Energy threshold for pruning
            unit: Unit for energy threshold 'hartree', 'kcal/mol' or 'kJ/mol'

        Returns:
            self: Self
        """
        # Convert threshold to Hartree.
        if unit.lower() == "hartree":
            pass
        if unit.lower() == "kcal/mol":
            threshold *= KCAL_TO_HARTREE
        elif unit.lower() == "kJ/mol":
            threshold *= KJ_TO_HARTREE

        # Prune conformers.
        energies = self.get_energies()
        energies -= np.min(energies)
        remove_list = np.where(energies > threshold)[0]

        for i in reversed(remove_list):
            del self.conformers[i]

        return self

    def prune_rmsd(
        self,
        thres: float = 0.35,
        include_hs: bool = False,
        symmetry: bool = False,
        method: str = "rdkit",
    ) -> "ConformerEnsemble":
        """Prune conformers based on RMSD.

        Args:
            thres: Threshold for RSMD pruning (Å)
            include_hs: Whether to include H atoms in RMSD calculation. Ignored for
                'obrms-iter' and 'obrms-batch' which only use heavy atoms.
            symmetry: Consider symmetry (requires connectivity matrix). Ignored for
                'obrms-iter' and 'obrms-batch' which always use symmetry.
            method: RMSD calculation method: 'obrms-batch', 'obrms-iter', 'openbabel',
                'rdkit' or 'spyrmsd'

        Returns:
            self: Self
        """
        # Select conformers to keep
        candidates = np.arange(len(self.conformers))
        keep_list = []

        # Prune conformers iteratively if possible to reduce number of RMSD
        # evaluations. For large systems obrms might be faster in batch.
        if method in ("obrms-iter", "openbabel", "spyrmsd", "rdkit"):
            while len(candidates) > 0:
                keeper = candidates[0]
                keep_list.append(keeper)
                rmsd = self.get_rmsd(
                    [keeper + 1],
                    candidates + 1,
                    include_hs=include_hs,
                    symmetry=symmetry,
                    method=method,
                )
                candidates = candidates[rmsd[0] > thres]
        elif method == "obrms-batch":
            rmsds = self.get_rmsd(
                include_hs=include_hs, symmetry=symmetry, method=method
            )
            working_array = rmsds
            while len(working_array) > 0:
                keeper = candidates[0]
                keep_list.append(keeper)
                rmsd = working_array[0]
                mask = rmsd > thres
                candidates = candidates[mask]
                working_array = working_array[mask, :][:, mask]

        # Update conformer list
        self.conformers = [
            conformer for i, conformer in enumerate(self.conformers) if i in keep_list
        ]

        return self

    def set_coordinates(
        self, conformer_coordinates: ArrayLike3D
    ) -> "ConformerEnsemble":
        """Set conformer coordinates.

        Args:
            conformer_coordinates: Conformer coordinates (Å)

        Returns:
            self: Self

        Raises:
            ValueError: When Number of conformer coordinates is different from number of
                conformers
        """
        conformer_coordinates = np.array(conformer_coordinates)
        if len(conformer_coordinates) != self.n_conformers:
            msg = (
                f"Number of coordinates ({len(conformer_coordinates)}) "
                f"!= number of conformers ({self.n_conformers})."
            )
            raise ValueError(msg)
        for conformer, coordinates in zip(self.conformers, conformer_coordinates):
            conformer.coordinates = np.array(coordinates)

        return self

    def set_degeneracies(self, degeneracies: ArrayLike1D) -> "ConformerEnsemble":
        """Set degeneracies.

        Args:
            degeneracies: Degeneracies

        Returns:
            self: Self

        Raises:
            ValueError: When number of degeneracies is different from number of
                conformers
        """
        degeneracies = np.array(degeneracies)
        if len(degeneracies) != self.n_conformers:
            msg = (
                f"Number of degeneracies ({len(degeneracies)}) "
                f"!= number of conformers ({self.n_conformers})."
            )
            raise ValueError(msg)
        for conformer, degeneracy in zip(self.conformers, degeneracies):
            conformer.degeneracy = degeneracy

        return self

    def set_energies(self, energies: ArrayLike1D) -> "ConformerEnsemble":
        """Set energies.

        Args:
            energies: Energy (a.u.)

        Returns:
            self: Self

        Raises:
            ValueError: When number of energies is different from number of
                conformers
        """
        energies = np.array(energies)
        if len(energies) != self.n_conformers:
            msg = (
                f"Number of energies ({len(energies)}) != number of "
                f"conformers ({self.n_conformers})."
            )
            raise ValueError(msg)
        for conformer, energy in zip(self.conformers, energies):
            conformer.energy = energy

        return self

    @requires_dependency([Import(module="rdkit.Chem", item="Descriptors")], globals())
    def set_multiplicity_from_mol(self) -> "ConformerEnsemble":
        """Sets multiplicity based on unpaired electrons in Mol object."""
        num_radical = Descriptors.NumRadicalElectrons(self.mol)

        # Assume maximum spin pairing
        if (num_radical % 2) == 0:
            multiplicity = 1
        else:
            multiplicity = 2

        self.multiplicity = multiplicity

        return self

    def set_properties(self, key: str, values: Iterable[float]) -> "ConformerEnsemble":
        """Set conformer properties.

        Args:
            key: Name of property
            values: Property values

        Returns:
            self: Self
        """
        for conformer, value in zip(self.conformers, values):
            conformer.properties[key] = value

        return self

    def sort(self) -> "ConformerEnsemble":
        """Sort conformers based on energy."""
        energies = [conformer.energy for conformer in self.conformers]
        if not all([isinstance(energy, float) for energy in energies]):
            raise ValueError("Not all conformers have energies.")
        energies = cast(List[float], energies)
        indices = np.argsort(energies)
        self.conformers = [self.conformers[i] for i in indices]

        return self

    def sp_qc_engine(
        self,
        ids: Optional[Sequence[int]] = None,
        program: str = "xtb",
        model: Optional[Dict[str, Any]] = None,
        keywords: Optional[Dict[str, Any]] = None,
        local_options: Optional[Dict[str, Any]] = None,
    ) -> "ConformerEnsemble":
        """Calculate conformer energies with QCEngine interface.

        Args:
            ids: Conformer indices (1-indexed). If None, all are calculated.
            program: QCEngine program
            model: QCEngine model
            keywords: QCEngine keywords
            local_options: QCEngine local options

        Returns:
            self: Self

        Raises:
            Exception: When trying to use RDKit with formal charges.
        """
        if np.any(self.formal_charges != 0) and (program.lower() == "rdkit"):
            raise Exception("QCEngine using RDKit does not work with charges.")

        # Set defaults
        if model is None:
            model = {"method": "GFN2-xTB"}
        if keywords is None:
            keywords = {}
        if local_options is None:
            local_options = {}
        if ids is not None:
            conformers = [self.conformers[i] for i in ids]
        else:
            conformers = self.conformers

        for conformer in conformers:
            energy = sp_qc_engine(
                self.elements,
                conformer.coordinates,
                charge=self.charge,
                multiplicity=self.multiplicity,
                connectivity_matrix=self.connectivity_matrix,
                program=program,
                model=model,
                keywords=keywords,
                local_options=local_options,
            )
            conformer.energy = energy

        return self

    def write_xyz(
        self,
        file: Union[str, PathLike],
        ids: Optional[Iterable[int]] = None,
        unit: str = "kcal/mol",
        relative: bool = True,
        separate: bool = False,
        n_decimals: int = 3,
    ) -> None:
        """Write conformers to xyz file.

        Args:
            file: Filename or path object. Needs filename if `separate=True`
            ids: Conformer indices (1-indexed)
            unit: Output unit for energies in xyz comment field: 'hartree', 'kcal/mol',
                'kJ/mol'
            relative: Whether to give energies relative to lowest energy conformer
            separate: Whether to write conformers to separate xyz files
            n_decimals: Number of decimals for energies

        Raises:
            TypeError: When separate=True and file is not str
        """
        if ids is None:
            ids_ = np.arange(len(self.conformers))
        else:
            ids_ = np.array(ids) - 1
        ids = ids_

        # Retrieve symbols, coordinates and energies
        symbols = convert_elements(self.elements, output="symbols")
        conformer_coordinates = self.get_coordinates()[ids]
        energies = self.get_relative_energies(unit=unit, relative=relative)[ids].round(
            n_decimals
        )

        # Write conformers
        if separate:
            if not isinstance(file, str):
                raise TypeError("file must be str when separate=True")
            for i, coordinates, energy in zip(ids, conformer_coordinates, energies):
                conf_filename = file.split(".")[0] + f"_{i + 1}.xyz"
                write_xyz(conf_filename, symbols, coordinates, comments=[energy])
        else:
            write_xyz(file, symbols, conformer_coordinates, comments=energies)

    def _add_conformers(
        self,
        conformer_coordinates: ArrayLike3D,
        energies: Optional[ArrayLike1D] = None,
        degeneracies: Optional[ArrayLike1D] = None,
        properties: Optional[Mapping[str, ArrayLike1D]] = None,
    ) -> None:
        conformer_coordinates = np.array(conformer_coordinates)
        n_conformers = len(conformer_coordinates)

        if energies is None:
            energies_ = np.full(n_conformers, np.nan)
        else:
            energies_ = np.array(energies)
        energies = energies_

        if degeneracies is None:
            degeneracies_ = np.ones(n_conformers)
        else:
            degeneracies_ = np.array(degeneracies)
        degeneracies = degeneracies_

        for coordinates, energy, degeneracy in zip(
            conformer_coordinates, energies, degeneracies
        ):
            conformer = Conformer(self.elements, coordinates, energy, degeneracy)
            self.conformers.append(conformer)
        if properties is not None:
            properties_ = {key: np.array(value) for key, value in properties.items()}
            for key, value in properties_.items():
                self.set_properties(key, value)

    @requires_executable(["obrms"])
    def _get_rmsd_obrms_batch(self, i_s: Array1D, j_s: Array1D) -> Array2D:
        """Calculate RMSD with obrms in batch mode.

        First calculates matrix of all pairwise RMSDs and then takes those of interest.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)

        Returns:
            rmsds: RMSDs (Å)
        """
        with tempfile.NamedTemporaryFile(suffix=".xyz") as ref_file:
            p_ref = Path(ref_file.name)
            self.write_xyz(p_ref, unit="hartree", relative=False)
            process = subprocess.run(
                f"obrms {p_ref.as_posix()} --cross " "--minimize".split(" "),
                capture_output=True,
            )
        result = np.genfromtxt(process.stdout.splitlines(), delimiter=",")
        # Reshape array to 2D if 1D
        if len(result.shape) == 1:
            result = result.reshape(1, -1)
        rmsds = result[:, 1:]

        rmsds: np.ndarray = rmsds[i_s - 1, :][:, j_s - 1]

        return rmsds

    @requires_executable(["obrms"])
    def _get_rmsd_obrms_iter(self, i_s: Array1D, j_s: Array1D) -> Array2D:
        """Calculate RMSD with obrms.

        Does iterative row-wise mode for heavy atoms and without symmetry.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)

        Returns:
            rmsds: Conformer RMSDs (Å)
        """
        rmsds = []
        for i in i_s:
            with tempfile.NamedTemporaryFile(
                suffix=".xyz"
            ) as ref_file, tempfile.NamedTemporaryFile(suffix=".xyz") as test_file:
                p_ref = Path(ref_file.name)
                p_test = Path(test_file.name)
                self.write_xyz(p_ref, ids=j_s)
                self.write_xyz(p_test, ids=[i])
                process = subprocess.run(
                    f"obrms {p_ref.as_posix()} "
                    f"{p_test.as_posix()} --minimize".split(" "),
                    capture_output=True,
                )
            row_rmsds = np.genfromtxt(process.stdout.splitlines(), usecols=(-1))
            rmsds.append(row_rmsds)
        rmsds = np.vstack(rmsds)

        return rmsds

    @requires_dependency([Import(module="openbabel.openbabel", alias="ob")], globals())
    def _get_rmsd_openbabel(
        self, i_s: Array1D, j_s: Array1D, include_hs: bool, symmetry: bool
    ) -> Array2D:
        """Calculate RMSD row-wise with openbabel python interface.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)
            include_hs: Whether to include H atoms in calculation
            symmetry: Whether to consider symmetry

        Returns:
            rmsds: Conformer RMSDs (Å)
        """
        rmsds = []
        for i in i_s:
            conformer_1 = self.conformers[i - 1]
            ob_mol_1 = _get_ob_mol(
                self.elements, conformer_1.coordinates, self.connectivity_matrix
            )
            align = ob.OBAlign(include_hs, symmetry)
            align.SetRefMol(ob_mol_1)
            rmsds_row = []
            for j in j_s:
                conformer_2 = self.conformers[j - 1]
                ob_mol_2 = _get_ob_mol(
                    self.elements, conformer_2.coordinates, self.connectivity_matrix
                )
                align.SetTargetMol(ob_mol_2)
                align.Align()
                rmsd = align.GetRMSD()
                rmsds_row.append(rmsd)
            rmsds.append(rmsds_row)
        rmsds = np.array(rmsds)

        return rmsds

    @requires_dependency(
        [
            Import(module="rdkit", item="Chem"),
            Import(module="rdkit.Chem", item="AllChem"),
        ],
        globals(),
    )
    def _get_rmsd_rdkit(self, i_s: Array1D, j_s: Array1D, include_hs: bool) -> Array2D:
        """Calculate RMSD row-wise with RDKit.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)
            include_hs: Whether to include H atoms in calculation

        Returns:
            rmsds: Conformer RMSDs (Å)
        """
        # Update mol object from conformers and make copy.
        self.update_mol()

        # Construct atom list for rmsd: heavy atoms or all atoms
        if include_hs:
            atom_ids = [atom.GetIdx() for atom in self.mol.GetAtoms()]
        else:
            atom_ids = [
                atom.GetIdx()
                for atom in self.mol.GetAtoms()
                if atom.GetAtomicNum() != 1
            ]

        # Calculated RMSD row-wise with RDKit
        rmsds = []
        conformers = list(self.mol.GetConformers())
        for i in i_s:
            ref_mol = Chem.Mol(self.mol)
            ref_conformer = conformers[i - 1]
            ref_mol.RemoveAllConformers()
            ref_mol.AddConformer(ref_conformer)
            for j in j_s:
                conformer = conformers[j - 1]
                ref_mol.AddConformer(conformer)
            rmsds_row: List[float] = []
            AllChem.AlignMolConformers(ref_mol, atomIds=atom_ids, RMSlist=rmsds_row)
            rmsds.append(rmsds_row)
        rmsds = np.array(rmsds)

        return rmsds

    @requires_dependency([Import("spyrmsd"), Import("spyrmsd.rmsd")], globals())
    def _get_rmsd_spyrmsd(
        self, i_s: Array1D, j_s: Array1D, include_hs: bool, symmetry: bool
    ) -> Array2D:
        """Calculate RMSD row-wise with spyrmsd.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)
            include_hs: Whether to include H atoms in calculation
            symmetry: Whether to consider symmetry

        Returns:
            rmsds: Conformer RMSDs (Å)
        """
        # Construct mask for H atoms
        if not include_hs:
            mask = self.elements != 1
        else:
            mask = np.ones(len(self.elements), dtype=bool)

        # Calculate RMSD row-wise with spyrmsd
        rmsds = []
        for i in i_s:
            ref_coordinates = self.conformers[i - 1].coordinates[mask]
            if symmetry:
                test_coordinates = [
                    self.conformers[j - 1].coordinates[mask] for j in j_s
                ]
                row_rmsds = spyrmsd.rmsd.symmrmsd(
                    ref_coordinates,
                    test_coordinates,
                    self.elements[mask],
                    self.elements[mask],
                    self.connectivity_matrix[mask, :][:, mask],
                    self.connectivity_matrix[mask, :][:, mask],
                    center=True,
                    minimize=True,
                )
            else:
                row_rmsds = []
                for j in j_s:
                    test_coordinates = self.conformers[j - 1].coordinates[mask]
                    rmsd = spyrmsd.rmsd.rmsd(
                        ref_coordinates,
                        test_coordinates,
                        self.elements[mask],
                        self.elements[mask],
                        center=True,
                        minimize=True,
                    )
                    row_rmsds.append(rmsd)
            rmsds.append(np.array(row_rmsds))
        rmsds = np.vstack(rmsds)

        return rmsds

    def update_mol(self) -> "ConformerEnsemble":
        """Update Mol object with conformers."""
        self.mol.RemoveAllConformers()
        conformer_coordinates = self.get_coordinates()
        _add_conformers_to_mol(self.mol, conformer_coordinates)

        return self

    def __copy__(self) -> "ConformerEnsemble":
        # Generate copy where conformers and mol object are shared with old
        # ensemble.
        cls = type(self)
        ce = cls(
            self.elements,
            connectivity_matrix=self.connectivity_matrix,
            charge=self.charge,
            multiplicity=self.multiplicity,
            formal_charges=self.formal_charges,
            ref_cip_label=self.ref_cip_label,
        )
        ce.mol = self.mol
        ce.conformers = self.conformers
        return ce

    def __deepcopy__(self, memo: Dict[int, object]) -> "ConformerEnsemble":
        # Generate copy where conformers and mol object are new
        cls = type(self)
        ce = cls(
            self.elements,
            conformer_coordinates=self.get_coordinates(),
            energies=self.get_energies(),
            connectivity_matrix=self.connectivity_matrix,
            degeneracies=self.get_degeneracies(),
            properties=self.get_properties(),
            charge=self.charge,
            multiplicity=self.multiplicity,
            formal_charges=self.formal_charges,
            ref_cip_label=self.ref_cip_label,
        )
        ce.mol = deepcopy(ce.mol, memo)
        return ce

    def __delitem__(self, index: int) -> None:
        del self.conformers[index]

    def __getitem__(
        self, index: Union[slice, numbers.Integral]
    ) -> Union["ConformerEnsemble", Conformer]:
        cls = type(self)
        if isinstance(index, slice):
            # Generate copy of ensemble with selected conformers.
            ce = copy(self)
            ce.conformers = ce.conformers[index]
            return ce
        elif isinstance(index, numbers.Integral):
            conformer = self.conformers[int(index)]
            return conformer
        else:
            raise TypeError(f"{cls.__name__} indices must be integers")

    def __len__(self) -> int:
        return len(self.conformers)

    def __repr__(self) -> str:
        n_conformers = len(self.conformers)
        return f"{self.__class__.__name__}({n_conformers!r} conformers)"


@requires_dependency(
    [
        Import(module="openbabel.openbabel", alias="ob"),
        Import(module="openbabel.pybel", alias="pybel"),
        Import("openbabel"),
    ],
    globals(),
)
def conformers_from_ob_ff(
    mol: Union[str, ob.OBMol],
    num_conformers: int = 30,
    ff: str = "MMFF94",
    method: str = "systematic",
    rings: bool = False,
) -> Tuple[Array1D, Array3D, Array2D, Array1D, ob.OBMol]:
    """Generates conformers based on the force field algorithm in OpenBabel.

    Follows the recipe of the command line script obabel --conformer:
    https://github.com/openbabel/openbabel/blob/master/src/ops/conformer.cpp
    If an OBMol object with 3D coordinates is given, the conformer search will
    start from that structure.

    Args:
        mol: Molecule either as SMILES string or OBMol object.
        num_conformers: Maximum number of conformers
        ff: Force field supported by OpenBabel
        method: 'fast', random', 'systematic' or 'weighted'
        rings: Sample ring torsions.

    Returns:
        elements: Elements as atomic numbers
        conformer_coordinates: Conformer coordinates (Å)
        connectivity_matrix: Connectivity matrix
        charges: Formal charges
        mol: OBMol object
    """
    # Create 3D structure if not given
    try:
        py_mol = pybel.readstring("smi", mol)
        py_mol.make3D()
        ob_mol = py_mol.OBMol
    except TypeError:
        ob_mol = mol
        py_mol = pybel.Molecule(ob_mol)

    # Set up FF
    ff = ob.OBForceField_FindForceField(ff)
    ff.EnableCutOff(True)
    ff.SetVDWCutOff(10.0)
    ff.SetElectrostaticCutOff(20.0)
    ff.SetUpdateFrequency(10)

    # Do conformer search
    if method == "systematic":
        ff.SystematicRotorSearch(10, rings)
    elif method == "fast":
        ff.FastRotorSearch(True)
    elif method == "random":
        ff.RandomRotorSearch(num_conformers, 10, rings)
    elif method == "weighted":
        ff.WeightedRotorSearch(num_conformers, 10, rings)
    ff.GetConformers(ob_mol)

    # Extract information
    (
        elements,
        conformer_coordinates,
        connectivity_matrix,
        charges,
    ) = _extract_from_ob_mol(ob_mol)

    return (elements, conformer_coordinates, connectivity_matrix, charges, ob_mol)


@requires_dependency(
    [
        Import(module="openbabel.openbabel", alias="ob"),
        Import(module="openbabel.pybel", alias="pybel"),
        Import("openbabel"),
    ],
    globals(),
)
def conformers_from_ob_ga(  # noqa: C901
    mol: Union[str, ob.OBMol],
    num_conformers: Optional[int] = None,
    num_children: Optional[int] = None,
    mutability: Optional[float] = None,
    convergence: Optional[int] = None,
    score: str = "rmsd",
    filter_method: str = "steric",
    cutoff: float = 0.8,
    vdw_factor: float = 0.5,
    check_hydrogens: bool = True,
) -> Tuple[Array1D, Array3D, Array2D, Array1D, ob.OBMol]:
    """Generates conformers based on the genetic algorithm in OpenBabel.

    Follows the recipe of the command line script obabel --conformer:
    https://github.com/openbabel/openbabel/blob/master/src/ops/conformer.cpp If an OBMol
    object with 3D coordinates is given, the conformer search will start from that
    structure.

    Args:
        mol: Molecule either as SMILES string or OBMol object
        num_conformers: Maximum number of conformers
        num_children: Number of children to generate for each parent
        mutability: Mutation frequency
        convergence: Number of identical generations before convergence is reached
        score: Scoring function: 'rmsd', 'min_rmsd', 'energy', 'min_energy'
        filter_method: Filtering algorithm: 'steric'
        cutoff: Absolute distance in Ånström below which atoms are considered to clash
        vdw_factor: Scale factor applied to van der Waals radii for detecting clashes
        check_hydrogens: Detect clashes with hydrogen atoms

    Returns:
        elements: Elements as atomic numbers
        conformer_coordinates: Conformer coordinates (Å)
        connectivity_matrix: Connectivity matrix
        charges: Formal charges
        mol: OBMol object
    """
    # Create 3D structure if not given
    try:
        py_mol = pybel.readstring("smi", mol)
        py_mol.make3D()
        ob_mol = py_mol.OBMol
    except TypeError:
        ob_mol = mol
        py_mol = pybel.Molecule(ob_mol)

    # Create search object and set parameters
    conf_search = ob.OBConformerSearch()
    conf_search.Setup(ob_mol)
    if num_conformers is not None:
        conf_search.SetNumConformers(num_conformers)
    if num_children is not None:
        conf_search.SetNumChildren(num_children)
    if convergence is not None:
        conf_search.SetConvergence(convergence)
    if mutability is not None:
        conf_search.SetMutability(mutability)
    if score is not None:
        # Scorers don't work with earlier versions of openbabel
        if not (parse_version(openbabel.__version__) > parse_version("3.1.0")):
            warnings.warn(
                "Scorer only works with openbabel version > 3.1.0. "
                "Proceeding without scorer."
            )
        else:
            if score == "rmsd":
                scorer = ob.OBRMSDConformerScore()
            if score == "min_rmsd":
                scorer = ob.OBMinimizingRMSDConformerScore()
            if score == "energy":
                scorer = ob.OBEnergyConformerScore()
            if score == "min_energy":
                scorer = ob.OBMinimizingEnergyConformerScore()
            conf_search.SetScore(scorer)
    if filter_method == "steric":
        # Filters don't work with earlier versions of openbabel
        if not (parse_version(openbabel.__version__) > parse_version("3.1.0")):
            warnings.warn(
                "Filter only works with openbabel version > 3.1.0. "
                "Proceeding without filter."
            )
        else:
            ob_filter = ob.OBStericConformerFilter(cutoff, vdw_factor, check_hydrogens)
            conf_search.SetFilter(ob_filter)

    # Do conformational search
    conf_search.Search()
    conf_search.GetConformers(ob_mol)

    # Extract information
    (
        elements,
        conformer_coordinates,
        connectivity_matrix,
        charges,
    ) = _extract_from_ob_mol(ob_mol)

    return (elements, conformer_coordinates, connectivity_matrix, charges, ob_mol)


@requires_dependency(
    [Import(module="rdkit", item="Chem"), Import(module="rdkit.Chem", item="AllChem")],
    globals(),
)
def conformers_from_rdkit(  # noqa: C901
    mol: Union[str, Chem.Mol],
    n_confs: Optional[int] = None,
    optimize: Optional[str] = "MMFF94",
    version: int = 2,
    small_rings: bool = True,
    macrocycles: bool = True,
    random_seed: Optional[int] = None,
    rmsd_thres: Optional[float] = 0.35,
    rmsd_symmetry: bool = False,
    n_threads: int = 1,
) -> Tuple[Array1D, Array3D, Optional[Array1D], Array2D, Array1D, Chem.Mol]:
    """Generates conformers for an RDKit mol object.

    Recipe based on J. Chem. Inf. Modeling 2012, 52, 1146.

    Args:
        mol: Molecule either as SMILES string or RDKit Mol object.
        n_confs: Number of conformers to generate. If None, a reasonable number will be
            set depending on the number of rotatable bonds.
        optimize: Force field used for conformer optimization: 'MMFF94', 'MMFF94s' or
            'UFF'. If None, conformers are not optimized.
        version: Version of the experimental torsion-angle preferences
        small_rings: Whether to impose small ring torsion angle preferences
        macrocycles: Whether to mpose macrocycle torsion angle preferences
        random_seed: Random seed for conformer generation
        rmsd_thres: Pruning RMSD threshold (Å)
        rmsd_symmetry: Whether to use symmetry for RMSD pruning
        n_threads: Number of threads

    Returns:
        elements: Atomic symbols
        conformer_coordinates: Coordinates for all conformers (Å)
        energies: Conformer energies (a.u.)
        connectivity_matrix: Connectivity matrix with bond orders
        charges: Formal charges
        mol: RDKit Mol object. Only returned when return_mol=True

    Raises:
        Exception: When force field not found
    """
    if optimize not in ("MMFF94", "MMFF94s", "UFF", None):
        raise Exception(
            f"Force field {optimize} not found. Choose one of" "MMFF94, MMFF94s, UFF."
        )
    # Generate mol object
    try:
        mol = Chem.MolFromSmiles(mol)
    except TypeError:
        pass
    mol = Chem.AddHs(mol)

    # If n_confs is not set, set number of conformers based on number of
    # rotatable bonds
    if not n_confs:
        n_rot_bonds = AllChem.CalcNumRotatableBonds(mol)

        if n_rot_bonds <= 7:
            n_confs = 50
        elif n_rot_bonds >= 8 and n_rot_bonds <= 12:
            n_confs = 200
        else:
            n_confs = 300

    # Generate conformers
    if rmsd_thres is not None:
        rdkit_prune_rmsd = rmsd_thres
    else:
        rdkit_prune_rmsd = -1
    if random_seed is not None:
        rdkit_random_seed = random_seed
    else:
        rdkit_random_seed = -1

    params = AllChem.EmbedParameters()
    params.useSymmetryForPruning = rmsd_symmetry
    params.randomSeed = rdkit_random_seed
    params.useSmallRingTorsions = small_rings
    params.useMacrocycleTorsions = macrocycles
    params.ETversion = version
    params.pruneRmsThresh = rdkit_prune_rmsd
    params.numThreads = n_threads
    AllChem.EmbedMultipleConfs(mol, n_confs, params)

    # Optimize with force fields
    results = None
    if optimize is not None:
        if optimize == "MMFF94":
            results = AllChem.MMFFOptimizeMoleculeConfs(
                mol, mmffVariant="MMFF94", numThreads=n_threads
            )
        elif optimize == "MMFF94s":
            results = AllChem.MMFFOptimizeMoleculeConfs(
                mol, mmffVariant="MMFF94s", numThreads=n_threads
            )
        if optimize == "UFF":
            results = AllChem.UFFOptimizeMoleculeConfs(mol, numThreads=n_threads)

    # Set energies
    if results is not None:
        energies = np.array([i[1] for i in results]) * KCAL_TO_HARTREE
    else:
        energies = None

    # Extract information from mol
    (elements, conformer_coordinates, connectivity_matrix, charges) = _extract_from_mol(
        mol
    )

    return (
        elements,
        conformer_coordinates,
        energies,
        connectivity_matrix,
        charges,
        mol,
    )


@requires_dependency(
    [
        Import(module="rdkit"),
        Import(module="rdkit", item="Chem"),
    ],
    globals(),
)
def _add_conformers_to_mol(mol: Chem.Mol, conformer_coordinates: ArrayLike3D) -> None:
    """Add conformers to RDKit Mol object.

    Args:
        mol: RDKit mol object
        conformer_coordinates: Conformer coordinates (Å)
    """
    conformer_coordinates = np.array(conformer_coordinates)
    if len(conformer_coordinates.shape) == 2:
        conformer_coordinates.reshape(-1, conformer_coordinates.shape[0], 3)

    for coordinates in conformer_coordinates:
        conformer = Chem.Conformer()
        for i, coord in enumerate(coordinates):
            point = rdkit.Geometry.Point3D(*coord)
            conformer.SetAtomPosition(i, point)
        mol.AddConformer(conformer, assignId=True)


@requires_dependency([Import(module="rdkit", item="Chem")], globals())
def _extract_from_mol(
    mol: Chem.Mol,
) -> Tuple[List[str], Array3D, Array2D, Array1D]:
    """Extract information from RDKit Mol object with conformers."""
    # Take out elements, coordinates and connectivity matrix
    elements = [atom.GetSymbol() for atom in mol.GetAtoms()]
    charges = np.array([atom.GetFormalCharge() for atom in mol.GetAtoms()])

    conformer_coordinates = []
    for conformer in mol.GetConformers():
        coordinates = conformer.GetPositions()
        conformer_coordinates.append(coordinates)
    conformer_coordinates = np.array(conformer_coordinates)

    connectivity_matrix = Chem.GetAdjacencyMatrix(mol, useBO=True)

    return elements, conformer_coordinates, connectivity_matrix, charges


@requires_dependency(
    [
        Import(module="openbabel.openbabel", alias="ob"),
        Import(module="openbabel.pybel", alias="pybel"),
        Import("openbabel"),
    ],
    globals(),
)
def _extract_from_ob_mol(
    ob_mol: ob.OBMol,
) -> Tuple[Array1D, Array3D, Array2D, Array1D]:
    """Extract information from Openbabel OBMol object with conformers."""
    py_mol = pybel.Molecule(ob_mol)
    elements = np.array([atom.atomicnum for atom in py_mol.atoms])
    charges = np.array([atom.formalcharge for atom in py_mol.atoms])

    n_atoms = len(py_mol.atoms)
    connectivity_matrix = np.zeros((n_atoms, n_atoms))
    for bond in ob.OBMolBondIter(ob_mol):
        i = bond.GetBeginAtomIdx() - 1
        j = bond.GetEndAtomIdx() - 1
        bo = bond.GetBondOrder()
        connectivity_matrix[i, j] = bo
        connectivity_matrix[j, i] = bo

    # Retrieve conformer coordinates
    conformer_coordinates = []
    for i in range(ob_mol.NumConformers()):
        ob_mol.SetConformer(i)
        coordinates = np.array([atom.coords for atom in py_mol.atoms])
        conformer_coordinates.append(coordinates)
    conformer_coordinates = np.array(conformer_coordinates)

    return elements, conformer_coordinates, connectivity_matrix, charges


@requires_dependency([Import(module="openbabel.openbabel", alias="ob")], globals())
def _get_ob_mol(
    elements: Union[Iterable[int], Iterable[str]],
    coordinates: ArrayLike2D,
    connectivity_matrix: ArrayLike2D,
    charges: Optional[ArrayLike1D] = None,
) -> ob.OBMol:
    """Generate OpenBabel OBMol object.

    Args:
        elements: Elements as atomic symbols or numbers.
        coordinates: Coordinates (Å)
        connectivity_matrix: Connectivity matrix with bond orders
        charges: Formal charges

    Returns:
        mol: OpenBabel OBMol object
    """
    elements = convert_elements(elements, output="numbers")
    if charges is None:
        charges_ = np.zeros(len(elements))
    else:
        charges_ = np.array(charges)
    charges = charges_

    connectivity_matrix = np.array(connectivity_matrix)
    coordinates = np.array(coordinates)

    mol = ob.OBMol()

    # Add atoms
    for element, coordinate, charge in zip(elements, coordinates, charges):
        a = mol.NewAtom()
        a.SetAtomicNum(int(element))
        a.SetVector(*coordinate)
        a.SetFormalCharge(int(charge))

    # Add bonds
    for i, j in zip(*np.tril_indices_from(connectivity_matrix)):
        if i != j:
            bo = connectivity_matrix[i, j]
            if bo != 0:
                mol.AddBond(int(i + 1), int(j + 1), int(bo))
    return mol


@requires_dependency([Import(module="rdkit", item="Chem")], globals())
def _get_rdkit_mol(
    elements: Union[Iterable[int], Iterable[str]],
    conformer_coordinates: ArrayLike3D,
    connectivity_matrix: ArrayLike2D,
    charges: Optional[ArrayLike1D] = None,
) -> Chem.Mol:
    _RDKIT_BOND_TYPES = {
        1.0: Chem.BondType.SINGLE,
        1.5: Chem.BondType.AROMATIC,
        2.0: Chem.BondType.DOUBLE,
        3.0: Chem.BondType.TRIPLE,
        4.0: Chem.BondType.QUADRUPLE,
        5.0: Chem.BondType.QUINTUPLE,
        6.0: Chem.BondType.HEXTUPLE,
    }
    elements = convert_elements(elements, output="symbols")

    if charges is None:
        charges_ = np.zeros(len(elements))
    else:
        charges_ = np.array(charges)
    charges = charges_
    conformer_coordinates = np.array(conformer_coordinates)
    connectivity_matrix = np.array(connectivity_matrix)

    mol = Chem.RWMol()

    # Add atoms
    for element, charge in zip(elements, charges):
        atom = Chem.Atom(element)
        atom.SetFormalCharge(int(charge))
        mol.AddAtom(atom)

    # Add bonds
    for i, j in zip(*np.tril_indices_from(connectivity_matrix)):
        if i != j:
            bo = connectivity_matrix[i, j]
            if bo != 0:
                bond_type = _RDKIT_BOND_TYPES[float(bo)]
                mol.AddBond(int(i), int(j), bond_type)

    # Add conformers
    _add_conformers_to_mol(mol, conformer_coordinates)

    mol = mol.GetMol()
    Chem.SanitizeMol(mol)

    return mol


def cli(smiles: str, generator: str = "rdkit") -> Any:
    """CLI for cone angle.

    Args:
        smiles: SMILES string
        generator: Confomer generator: 'ob-ff', 'ob-ga' or 'rdkit'

    Returns:
        Partially instantiated class
    """
    if generator == "rdkit":
        return functools.partial(ConformerEnsemble.from_rdkit, smiles)
    elif generator == "ob-ff":
        return functools.partial(ConformerEnsemble.from_ob_ff, smiles)
    elif generator == "ob-ga":
        return functools.partial(ConformerEnsemble.from_ob_ga, smiles)



"""Plotting functions."""

import typing
from typing import Optional, Sequence

import numpy as np

from morfeus.utils import Import, requires_dependency

if typing.TYPE_CHECKING:
    import pyvista as pv
    import vtk


@requires_dependency([Import(module="pyvista", alias="pv")], globals())
def get_drawing_arrow(
    start: Optional[Sequence[float]] = None,
    direction: Optional[Sequence[float]] = None,
    length: float = 1,
    shaft_radius: float = 0.05,
    shaft_resolution: int = 20,
    tip_length: float = 0.25,
    tip_radius: float = 0.1,
    tip_resolution: int = 20,
) -> "pv.MultiBlock":
    """Creates PyVista 3D arrow from cone and cylider.

    Args:
        start: Starting point (Å)
        direction: Direction vector (Å)
        length: Length (Å)
        shaft_radius: Shaft radius (Å)
        shaft_resolution: Shaft resolution
        tip_length: Tip length (Å)
        tip_radius:: Tip radius (Å)
        tip_resolution: Tip resoluation

    Returns:
        arrow: 3D arrow
    """
    # Set start and direction
    if start is None:
        start = [0, 0, 0]
    if direction is None:
        direction = [1, 0, 0]
    start = np.array(start)
    direction = np.array(direction) / np.linalg.norm(direction)

    # Create cylinder
    cylinder_length = length - tip_length
    cylinder_center = start + length * direction / 2
    cyl = pv.Cylinder(
        center=cylinder_center,
        direction=direction,
        radius=shaft_radius,
        height=cylinder_length,
        resolution=shaft_resolution,
    )

    # Create cone
    cone_center = start + (cylinder_length + tip_length / 2) * direction
    cone = get_drawing_cone(
        center=cone_center,
        direction=direction,
        radius=tip_radius,
        height=tip_length,
        resolution=tip_resolution,
    )

    # Assemble to arrow
    arrow = pv.MultiBlock()
    arrow.append(cyl)
    arrow.append(cone)

    return arrow


@requires_dependency([Import(module="pyvista", alias="pv"), Import("vtk")], globals())
def get_drawing_cone(
    center: Optional[Sequence[float]] = None,
    direction: Optional[Sequence[float]] = None,
    height: float = 1.0,
    radius: Optional[float] = None,
    capping: bool = True,
    angle: Optional[float] = None,
    resolution: int = 6,
) -> "pv.PolyData":
    """Create a cone.

    Copy from the PyVista code.

    Args:
        center: Center in [x, y, z]. middle of the axis of the cone.
        direction: Direction vector in [x, y, z]. orientation vector of the cone.
        height: Height along the cone in its specified direction.
        radius: Base radius of the cone
        capping: Turn on/off whether to cap the base of the cone with a polygon.
        angle: The angle degrees between the axis of the cone and a generatrix.
        resolution: Number of facets used to represent the cone

    Returns:
        cone: 3D cone

    Raises:
        Exception: When both radius and angle are specified.
    """
    if center is None:
        center = [0.0, 0.0, 0.0]
    if direction is None:
        direction = [1.0, 0.0, 0.0]

    src = vtk.vtkConeSource()
    src.SetCapping(capping)
    src.SetDirection(direction)
    src.SetCenter(center)
    src.SetHeight(height)
    if angle and radius:
        raise Exception("Both radius and angle specified. They are mutually exclusive.")
    elif angle and not radius:
        src.SetAngle(angle)
    elif not angle and radius:
        src.SetRadius(radius)
    elif not angle and not radius:
        src.SetRadius(0.5)
    src.SetResolution(resolution)
    src.Update()
    cone = pv.wrap(src.GetOutput())

    return cone



"""Internal calculators."""

import typing
from typing import Dict, Iterable, List, Union

import numpy as np
from scipy.spatial.distance import cdist

from morfeus.d3_data import c6_reference_data, r2_r4
from morfeus.data import ANGSTROM, BOHR, EV, HARTREE
from morfeus.geometry import Atom
from morfeus.typing import Array1D, ArrayLike2D
from morfeus.utils import convert_elements, get_radii, Import, requires_dependency

if typing.TYPE_CHECKING:
    import ase
    from dftd4.calculators import D3_model, D4_model
    from dftd4.utils import extrapolate_c_n_coeff


@requires_dependency(
    [
        Import("ase"),
        Import(module="dftd4.calculators", item="D3_model"),
        Import(module="dftd4.utils", item="extrapolate_c_n_coeff"),
    ],
    globals(),
)
class D3Grimme:
    """Calculates D3-like Cᴬᴬ coefficients with dftd4.

    Args:
        elements : Elements as atomic symbols or numbers
        coordinates : Coordinates (Å)
        order: Maximum order for the CN coefficients.

    Attributes:
        c_n_coefficients: Cᴬᴬ coefficients (a.u.)
        polarizabilities: Atomic polarizabilities (a.u.)
    """

    c_n_coefficients: Dict[int, Array1D]
    polarizabilities: Array1D
    _atoms: List["ase.Atoms"]

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        order: int = 6,
    ) -> None:
        # Convert elements to atomic numbers
        elements = convert_elements(elements, output="numbers")

        # Do calculation
        atoms = ase.Atoms(numbers=elements, positions=coordinates)
        calc = D3_model(energy=False, forces=False)
        polarizabilities = (
            calc.get_property("polarizibilities", atoms=atoms) * (ANGSTROM / BOHR) ** 3
        )
        c6_coefficients_all = (
            calc.get_property("c6_coefficients", atoms=atoms)
            * EV
            / HARTREE
            * (ANGSTROM / BOHR) ** 6
        )
        c6_coefficients = np.diag(c6_coefficients_all)
        c_n_coefficients = {}
        c_n_coefficients[6] = c6_coefficients

        # Extrapolate
        for i in range(8, order + 1, 2):
            c_n_coefficients[i] = np.array(
                [
                    extrapolate_c_n_coeff(c6, element, element, i)
                    for c6, element in zip(c6_coefficients, elements)
                ]
            )

        # Store attributes
        self.polarizabilities = polarizabilities
        self.c_n_coefficients = c_n_coefficients
        self._atoms = atoms

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


@requires_dependency(
    [
        Import("ase"),
        Import(module="dftd4.calculators", item="D4_model"),
        Import(module="dftd4.utils", item="extrapolate_c_n_coeff"),
    ],
    globals(),
)
class D4Grimme:
    """Calculates D4 Cᴬᴬ dispersion coefficients with dftd4.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        order: Maximum order for the CN coefficients.
        charge: Molecular charge

    Attributes:
        c_n_coefficients: Cᴬᴬ coefficients (a.u.)
        charges: Atomic charges.
        polarizabilities: Atomic polarizabilities (a.u.)
    """

    c_n_coefficients: Dict[int, Array1D]
    charges: Array1D
    polarizabilities: Array1D
    _atoms: List["ase.Atoms"]

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        order: int = 8,
        charge: int = 0,
    ) -> None:
        # Convert elements to atomic numbers
        elements = convert_elements(elements, output="numbers")

        # Set up atoms object
        charges = np.zeros(len(elements))
        charges[0] = charge
        atoms = atoms = ase.Atoms(
            numbers=elements, positions=coordinates, charges=charges
        )

        # Do calculation
        calc = D4_model(energy=False, forces=False)
        polarizabilities = (
            calc.get_property("polarizibilities", atoms=atoms) * (ANGSTROM / BOHR) ** 3
        )
        charges = calc.get_property("charges", atoms=atoms)
        c6_coefficients_all = (
            calc.get_property("c6_coefficients", atoms=atoms)
            * EV
            / HARTREE
            * (ANGSTROM / BOHR) ** 6
        )
        c6_coefficients = np.diag(c6_coefficients_all)
        c_n_coefficients = {}
        c_n_coefficients[6] = c6_coefficients

        # Extrapolate
        for i in range(8, order + 1, 2):
            c_n_coefficients[i] = np.array(
                [
                    extrapolate_c_n_coeff(c6, element, element, i)
                    for c6, element in zip(c6_coefficients, elements)
                ]
            )

        # Store attributes
        self.polarizabilities = polarizabilities
        self.charges = charges
        self.c_n_coefficients = c_n_coefficients
        self._atoms = atoms

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


class D3Calculator:
    """Calculates D3 Cᴬᴬ coefficients.

    Procedure as described in J. Chem. Phys. 2010, 132, 154104.

    Args:
        elements : Elements as atomic symbols or numbers
        coordinates : Coordinates (Å)
        order: Maximum order for the CN coefficients.

    Attributes:
        c_n_coefficients: Cᴬᴬ coefficients (a.u.)
        coordination_numbers: Atomic coordination numbers.
    """

    c_n_coefficients: Dict[int, Array1D]
    coordination_numbers: Array1D
    _atoms: List[Atom]

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        order: int = 8,
    ) -> None:
        # Convert elements to atomic numbers
        elements = convert_elements(elements, output="numbers")
        coordinates = np.array(coordinates)

        # Load the covalent radii
        radii = get_radii(elements, radii_type="pyykko")

        # Set up the atoms objects
        atoms = []
        for i, (element, coordinate, radius) in enumerate(
            zip(elements, coordinates, radii), start=1
        ):
            atom = Atom(element, coordinate, radius, i)
            atoms.append(atom)
        self._atoms = atoms

        # Calculate the coordination numbers according to Grimme's recipe.
        k_1 = 16
        k_2 = 4 / 3
        k_3 = 4
        for cn_atom in atoms:
            # Get coordinates and radii of all other atoms
            other_coordinates = np.array(
                [atom.coordinates for atom in atoms if atom is not cn_atom]
            )
            other_radii = np.array(
                [atom.radius for atom in atoms if atom is not cn_atom]
            )

            # Calculate distances
            dists = cdist(
                np.array(cn_atom.coordinates).reshape(-1, 3),
                other_coordinates.reshape(-1, 3),
            )

            # Calculate coordination number
            coordination_number = np.sum(
                1
                / (
                    1
                    + np.exp(-k_1 * (k_2 * (cn_atom.radius + other_radii) / dists - 1))
                )
            )
            cn_atom.coordination_number = coordination_number

        # Calculate the C_N coefficients
        c_n_coefficients: Dict[int, List[float]] = {
            i: [] for i in range(6, order + 1, 2)
        }
        for atom in atoms:
            # Get the reference data for atom
            reference_data = c6_reference_data[atom.element]
            cn = atom.coordination_number

            # Take out the coordination numbers and c6(aa) values
            c_6_ref = reference_data[:, 0]
            cn_1 = reference_data[:, 1]
            cn_2 = reference_data[:, 2]

            # Calculate c6  and c8 according to the recipe
            r = (cn - cn_1) ** 2 + (cn - cn_2) ** 2
            L = np.exp(-k_3 * r)
            W = np.sum(L)
            Z = np.sum(c_6_ref * L)
            c_6 = Z / W

            temp_coefficients: Dict[int, float] = {
                i: np.nan for i in range(6, order + 1, 2)
            }
            for i in range(6, order + 1, 2):
                if i == 6:
                    temp_coefficients[i] = c_6
                if i == 8:
                    c_8 = 3 * c_6 * r2_r4[atom.element] ** 2
                    temp_coefficients[i] = c_8
                elif i == 10:
                    c_10 = 49.0 / 40.0 * c_8 ** 2 / c_6
                    temp_coefficients[i] = c_10
                elif i > 10:
                    c_n = (
                        temp_coefficients[i - 6]
                        * (temp_coefficients[i - 2] / temp_coefficients[i - 4]) ** 3
                    )
                    temp_coefficients[i] = c_n
            for key, value in temp_coefficients.items():
                c_n_coefficients[key].append(value)

        # Set up attributes
        coordination_numbers = np.array(
            [atom.coordination_number for atom in self._atoms]
        )
        c_n_coefficients = {
            key: np.array(value) for key, value in c_n_coefficients.items()
        }

        self._atoms = atoms
        self.c_n_coefficients = c_n_coefficients
        self.coordination_numbers = coordination_numbers

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"



"""Buried volume code."""

import copy
import functools
import itertools
import math
import typing
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, Union
import warnings

import numpy as np
import scipy.spatial

from morfeus.data import jmol_colors
from morfeus.geometry import Atom, kabsch_rotation_matrix, rotate_coordinates, Sphere
from morfeus.io import read_geometry
from morfeus.sasa import SASA
from morfeus.typing import Array1D, Array2D, ArrayLike1D, ArrayLike2D
from morfeus.utils import convert_elements, get_radii, Import, requires_dependency

if typing.TYPE_CHECKING:
    from matplotlib.colors import hex2color
    import matplotlib.pyplot as plt
    from pyvista import BackgroundPlotter
    import pyvista as pv

# Quadrant and octant signs taken from
# https://en.wikipedia.org/wiki/Octant_(solid_geometry)
QUADRANT_SIGNS: Dict[int, str] = {
    1: "+,+",
    2: "-,+",
    3: "-,-",
    4: "+,-",
}
"""Coventional signs for quadrants."""

OCTANT_SIGNS: Dict[int, str] = {
    0: "+,+,+",
    1: "-,+,+",
    3: "+,-,+",
    2: "-,-,+",
    7: "+,+,-",
    6: "-,+,-",
    4: "+,-,-",
    5: "-,-,-",
}
"""Conventional signs for octants."""

QUADRANT_NAMES: Dict[int, str] = {
    1: "NE",
    2: "NW",
    3: "SW",
    4: "SE",
}
"""Conventional names for quadrants."""

# Maps octants to quadrants
QUADRANT_OCTANT_MAP: Dict[int, Tuple[int, int]] = {
    1: (0, 7),
    2: (1, 6),
    3: (2, 5),
    4: (3, 4),
}
"""Map from quadrants to octants."""


class BuriedVolume:
    """Performs and stores the results of a buried volume calculation.

    Algorithm similar as to described in Organometallics 2016, 35, 2286.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        metal_index: Index of metal atom (1-indexed)
        excluded_atoms: Indices of atoms to exclude (1-indexed). Metal atom is always
            excluded and does not have to be given here.
        radii: vdW radii (Å)
        include_hs: Whether to include H atoms in the calculation
        radius: Radius of sphere (Å)
        radii_type: Type of radii to use: 'alvarez', 'bondi', 'crc' or 'truhlar'
        radii_scale: Scaling factor for radii
        density: Volume per point in the sphere (Å³)
        z_axis_atoms: Atom indices for deciding orientation of z axis (1-indexed)
        xy_plane_atoms: Atom indices for deciding orientation of xz plane (1-indexed)

    Attributes:
        buried_volume: Buried volume of sphere (Å³)
        distal_volume: Distal volume of ligand (Å³)
        fraction_buried_volume: Fraction buried volume of sphere
        free_volume: Free volume of sphere (Å³)
        octants: Results for octant analysis
        quadrants: Results for quadrant analysis
    """

    buried_volume: float
    distal_volume: float
    fraction_buried_volume: float
    free_volume: float
    molecular_volume: float
    octants: Dict[str, Dict[int, float]]
    quadrants: Dict[str, Dict[int, float]]
    _all_coordinates: Array2D
    _atoms: List[Atom]
    _buried_points: Array1D
    _density: float
    _excluded_atoms: Set[int]
    _free_points: Array1D
    _octant_limits: Dict[
        int, Tuple[Tuple[float, float], Tuple[float, float], Tuple[float, float]]
    ]
    _sphere: Sphere

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        metal_index: int,
        excluded_atoms: Optional[Sequence[int]] = None,
        radii: Optional[ArrayLike1D] = None,
        include_hs: bool = False,
        radius: float = 3.5,
        radii_type: str = "bondi",
        radii_scale: float = 1.17,
        density: float = 0.001,
        z_axis_atoms: Optional[Sequence[int]] = None,
        xz_plane_atoms: Optional[Sequence[int]] = None,
    ) -> None:
        # Get center and and reortient coordinate system
        coordinates = np.array(coordinates)
        center = coordinates[metal_index - 1]
        coordinates -= center

        if excluded_atoms is None:
            excluded_atoms = []
        excluded_atoms = set(excluded_atoms)

        if metal_index not in excluded_atoms:
            excluded_atoms.add(metal_index)

        if z_axis_atoms is not None and xz_plane_atoms is not None:
            z_axis_coordinates = coordinates[np.array(z_axis_atoms) - 1]
            z_point = np.mean(z_axis_coordinates, axis=0)

            xz_plane_coordinates = coordinates[np.array(xz_plane_atoms) - 1]
            xz_point = np.mean(xz_plane_coordinates, axis=0)

            v_1 = z_point - center
            v_2 = xz_point - center
            v_3 = np.cross(v_2, v_1)
            real = np.vstack([v_1, v_3])
            real /= np.linalg.norm(real, axis=1).reshape(-1, 1)
            ref_1 = np.array([0.0, 0.0, -1.0])
            ref_2 = np.array([0.0, 1.0, 0.0])
            ref = np.vstack([ref_1, ref_2])
            R = kabsch_rotation_matrix(real, ref, center=False)
            coordinates = (R @ coordinates.T).T
        elif z_axis_atoms is not None:
            z_axis_coordinates = coordinates[np.array(z_axis_atoms) - 1]
            z_point = np.mean(z_axis_coordinates, axis=0)
            v_1 = z_point - center
            v_1 = v_1 / np.linalg.norm(v_1)
            coordinates = rotate_coordinates(coordinates, v_1, np.array([0, 0, -1]))

        self._z_axis_atoms = z_axis_atoms
        self._xz_plane_atoms = xz_plane_atoms

        # Save density and coordinates for steric map plotting.
        self._density = density
        self._all_coordinates = coordinates

        # Converting element ids to atomic numbers if the are symbols
        elements = convert_elements(elements, output="numbers")

        # Getting radii if they are not supplied
        if radii is None:
            radii = get_radii(elements, radii_type=radii_type, scale=radii_scale)
        radii = np.array(radii)

        # Get list of atoms as Atom objects
        atoms = []
        for i, (element, radius_, coord) in enumerate(
            zip(elements, radii, coordinates), start=1
        ):
            if i in excluded_atoms:
                continue
            elif (not include_hs) and element == 1:
                continue
            else:
                atom = Atom(element, coord, radius_, i)
                atoms.append(atom)

        # Set variables for outside access and function access.
        self._atoms = atoms
        self._excluded_atoms = set(excluded_atoms)

        # Compute buried volume
        self._compute_buried_volume(center=center, radius=radius, density=density)

    def octant_analysis(self) -> "BuriedVolume":
        """Perform octant analysis of the buried volume."""
        # Set up limits depending on the sphere radius
        lim = self._sphere.radius
        octant_limits = {
            0: ((0, lim), (0, lim), (0, lim)),
            1: ((-lim, 0), (0, lim), (0, lim)),
            3: ((0, lim), (-lim, 0), (0, lim)),
            2: ((-lim, 0), (-lim, 0), (0, lim)),
            7: ((0, lim), (0, lim), (-lim, 0)),
            6: ((-lim, 0), (0, lim), (-lim, 0)),
            4: ((0, lim), (-lim, 0), (-lim, 0)),
            5: ((-lim, 0), (-lim, 0), (-lim, 0)),
        }

        # Calculated volume for each octant.
        octant_volume = self._sphere.volume / 8

        # Do octant analysis
        percent_buried_volume = {}
        buried_volume = {}
        free_volume = {}
        for name, limits in octant_limits.items():
            buried_points = self._buried_points[
                np.logical_and.reduce(
                    [
                        self._buried_points[:, 0] > limits[0][0],
                        self._buried_points[:, 0] < limits[0][1],
                        self._buried_points[:, 1] > limits[1][0],
                        self._buried_points[:, 1] < limits[1][1],
                        self._buried_points[:, 2] > limits[2][0],
                        self._buried_points[:, 2] < limits[2][1],
                    ]
                )
            ]
            free_points = self._free_points[
                np.logical_and.reduce(
                    [
                        self._free_points[:, 0] > limits[0][0],
                        self._free_points[:, 0] < limits[0][1],
                        self._free_points[:, 1] > limits[1][0],
                        self._free_points[:, 1] < limits[1][1],
                        self._free_points[:, 2] > limits[2][0],
                        self._free_points[:, 2] < limits[2][1],
                    ]
                )
            ]
            fraction_buried = len(buried_points) / (
                len(buried_points) + len(free_points)
            )
            percent_buried_volume[name] = fraction_buried * 100
            buried_volume[name] = fraction_buried * octant_volume
            free_volume[name] = (1 - fraction_buried) * octant_volume
        self.octants = {
            "percent_buried_volume": percent_buried_volume,
            "buried_volume": buried_volume,
            "free_volume": free_volume,
        }

        # Do quadrant analysis
        percent_buried_volume = {}
        buried_volume = {}
        free_volume = {}
        for name, octants in QUADRANT_OCTANT_MAP.items():
            percent_buried_volume[name] = (
                sum(
                    [
                        self.octants["percent_buried_volume"][octant]
                        for octant in octants
                    ]
                )
                / 2
            )
            buried_volume[name] = sum(
                [self.octants["buried_volume"][octant] for octant in octants]
            )
            free_volume[name] = sum(
                [self.octants["free_volume"][octant] for octant in octants]
            )
        self.quadrants = {
            "percent_buried_volume": percent_buried_volume,
            "buried_volume": buried_volume,
            "free_volume": free_volume,
        }

        self._octant_limits = octant_limits

        return self

    def _compute_buried_volume(
        self, center: ArrayLike1D, radius: float, density: float
    ) -> None:
        """Compute buried volume."""
        center = np.array(center)
        # Construct sphere at metal center
        sphere = Sphere(
            center, radius, method="projection", density=density, filled=True
        )

        # Prune sphere points which are within vdW radius of other atoms.
        tree = scipy.spatial.cKDTree(
            sphere.points, compact_nodes=False, balanced_tree=False
        )
        mask = np.zeros(len(sphere.points), dtype=bool)
        for atom in self._atoms:
            if atom.radius + sphere.radius > np.linalg.norm(atom.coordinates):
                to_prune = tree.query_ball_point(atom.coordinates, atom.radius)
                mask[to_prune] = True
        buried_points = sphere.points[mask, :]
        free_points = sphere.points[np.invert(mask), :]

        # Calculate buried_volume
        self.fraction_buried_volume = len(buried_points) / len(sphere.points)
        self.buried_volume = sphere.volume * self.fraction_buried_volume
        self.free_volume = sphere.volume - self.buried_volume
        self._sphere = sphere
        self._buried_points = buried_points
        self._free_points = free_points

    def compute_distal_volume(
        self, method: str = "sasa", octants: bool = False, sasa_density: float = 0.01
    ) -> "BuriedVolume":
        """Computes the distal volume.

        Uses either SASA or Buried volume with large radius to calculate the molecular
        volume.

        Args:
            method: Method to get total volume: 'buried_volume' or 'sasa'
            octants: Whether to compute distal volume for quadrants and octants.
                Requires method='buried_volume'
            sasa_density: Density of points on SASA surface. Ignored unless
                method='sasa'

        Returns:
            self: Self

        Raises:
            ValueError: When method is not specified correctly.
        """
        loop_coordinates: List[np.ndarray]
        # Use SASA to calculate total volume of the molecule
        if method == "sasa":
            # Calculate total volume
            elements: List[int] = []
            loop_coordinates = []
            radii: List[float] = []
            for atom in self._atoms:
                elements.append(atom.element)
                loop_coordinates.append(atom.coordinates)
                radii.append(atom.radius)
            coordinates = np.vstack(loop_coordinates)
            sasa = SASA(
                elements,
                coordinates,
                radii=radii,
                probe_radius=0.0,
                density=sasa_density,
            )
            self.molecular_volume = sasa.volume

            # Calculate distal volume
            self.distal_volume = self.molecular_volume - self.buried_volume
        elif method == "buried_volume":
            if octants is True and self.octants is None:
                raise ValueError("Needs octant analysis.")

            # Save the values for the old buried volume calculation
            temp_bv = copy.deepcopy(self)

            # Determine sphere radius to cover the whole molecule
            loop_coordinates = []
            radii = []
            for atom in self._atoms:
                loop_coordinates.append(atom.coordinates)
                radii.append(atom.radius)
            coordinates = np.vstack(loop_coordinates)
            distances = scipy.spatial.distance.cdist(
                self._sphere.center.reshape(1, -1), coordinates
            )
            new_radius = np.max(distances + radii) + 0.5

            # Compute the distal volume
            temp_bv._compute_buried_volume(
                center=self._sphere.center,
                radius=new_radius,
                density=self._sphere.density,
            )
            self.molecular_volume = temp_bv.buried_volume
            self.distal_volume = self.molecular_volume - self.buried_volume
            if octants is True:
                temp_bv.octant_analysis()

                # Octant analysis
                distal_volume = {}
                molecular_volume = {}
                for name in self.octants["buried_volume"].keys():
                    molecular_volume[name] = temp_bv.octants["buried_volume"][name]
                    distal_volume[name] = (
                        temp_bv.octants["buried_volume"][name]
                        - self.octants["buried_volume"][name]
                    )
                self.octants["distal_volume"] = distal_volume
                self.octants["molecular_volume"] = molecular_volume

                # Quadrant analyis
                distal_volume = {}
                molecular_volume = {}
                for name, octants_ in QUADRANT_OCTANT_MAP.items():
                    distal_volume[name] = sum(
                        [self.octants["distal_volume"][octant] for octant in octants_]
                    )
                    molecular_volume[name] = sum(
                        [
                            self.octants["molecular_volume"][octant]
                            for octant in octants_
                        ]
                    )
                self.quadrants["distal_volume"] = distal_volume
                self.quadrants["molecular_volume"] = molecular_volume
        else:
            raise ValueError(f"Method {method} is not valid.")

        return self

    @requires_dependency([Import(module="matplotlib.pyplot", alias="plt")], globals())
    def plot_steric_map(  # noqa: C901
        self,
        filename: Optional[str] = None,
        levels: float = 150,
        grid: int = 100,
        all_positive: bool = True,
        cmap: str = "viridis",
    ) -> None:
        """Plots a steric map as in the original article.

        Args:
            filename: Name of file for saving the plot.
            levels: Number of levels in the contour plot
            grid: Number of points along each axis of plotting grid
            all_positive: Whether to plot only positive values
            cmap: Matplotlib colormap for contour plot

        Raises:
            ValueError: When z-axis atoms not present
        """
        if self._z_axis_atoms is None:
            raise ValueError("Must give z-axis atoms when instantiating BuriedVolume.")
        # Set up coordinates
        atoms = self._atoms
        center = np.array(self._sphere.center)
        all_coordinates = self._all_coordinates
        coordinates = np.array([atom.coordinates for atom in atoms])

        # Translate coordinates
        all_coordinates -= center
        coordinates -= center
        center -= center

        # Get vector to midpoint of z-axis atoms
        z_axis_coordinates = all_coordinates[np.array(self._z_axis_atoms) - 1]
        point = np.mean(z_axis_coordinates, axis=0)
        vector = point - center
        vector = vector / np.linalg.norm(vector)

        # Rotate coordinate system
        coordinates = rotate_coordinates(coordinates, vector, np.array([0, 0, -1]))

        # Make grid
        r = self._sphere.radius
        x_ = np.linspace(-r, r, grid)
        y_ = np.linspace(-r, r, grid)

        # Calculate z values
        z = []
        for line in np.dstack(np.meshgrid(x_, y_)).reshape(-1, 2):
            if np.linalg.norm(line) > r:
                z.append(np.nan)
                continue
            x = line[0]
            y = line[1]
            z_list = []
            for i, atom in enumerate(atoms):
                # Check if point is within reach of atom.
                x_s = coordinates[i, 0]
                y_s = coordinates[i, 1]
                z_s = coordinates[i, 2]
                test = atom.radius ** 2 - (x - x_s) ** 2 - (y - y_s) ** 2
                if test >= 0:
                    z_atom = math.sqrt(test) + z_s
                    z_list.append(z_atom)
            # Take point which is furthest along z axis
            if z_list:
                z_max = max(z_list)
                # Test if point is inside the sphere. Points with positive z
                # values are included by default anyway in accordance to
                # article
                if all_positive:
                    if z_max < 0:
                        if np.linalg.norm(np.array([x, y, z_max])) >= r:
                            z_max = np.nan
                else:
                    if np.linalg.norm(np.array([x, y, z_max])) >= r:
                        z_max = np.nan
            else:
                z_max = np.nan
            z.append(z_max)

        # Create interaction surface
        z = np.array(z).reshape(len(x_), len(y_))

        # Plot surface
        fig, ax = plt.subplots()
        cf = ax.contourf(x_, y_, z, levels, cmap=cmap)
        circle = plt.Circle((0, 0), r, fill=False)
        ax.add_patch(circle)
        plt.xlabel("x (Å)")
        plt.ylabel("y (Å)")
        cf.set_clim(-r, r)
        c_bar = fig.colorbar(cf)
        c_bar.set_label("z(Å)")
        ax.set_aspect("equal", "box")

        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def print_report(self) -> None:
        """Prints a report of the buried volume."""
        print("V_bur (%):", round(self.fraction_buried_volume * 100, 1))

    @requires_dependency(
        [
            Import(module="matplotlib.colors", item="hex2color"),
            Import(module="pyvista", alias="pv"),
            Import(module="pyvistaqt", item="BackgroundPlotter"),
        ],
        globals(),
    )
    def draw_3D(
        self,
        atom_scale: float = 1,
        background_color: str = "white",
        buried_color: str = "tomato",
        free_color: str = "steelblue",
        opacity: float = 0.5,
        size: float = 5,
    ) -> None:
        """Draw a the molecule with the buried and free points.

        Args:
            atom_scale: Scaling factor for atom size
            background_color: Background color for plot
            buried_color: Color of buried points
            free_color: Color of free points
            opacity: Point opacity
            size: Point size
        """
        # Set up plotter
        p = BackgroundPlotter()
        p.set_background(background_color)

        # Draw molecule
        for atom in self._atoms:
            color = hex2color(jmol_colors[atom.element])
            radius = atom.radius * atom_scale
            sphere = pv.Sphere(center=list(atom.coordinates), radius=radius)
            p.add_mesh(sphere, color=color, opacity=1, name=str(atom.index))

        # Add buried points
        p.add_points(
            self._buried_points, color=buried_color, opacity=opacity, point_size=size
        )

        # Add free points
        p.add_points(
            self._free_points, color=free_color, opacity=opacity, point_size=size
        )

        if hasattr(self, "_octant_limits"):
            for name, limits_ in self._octant_limits.items():
                limits = tuple(itertools.chain(*limits_))
                box = pv.Box(limits)
                p.add_mesh(box, style="wireframe")
                x = np.array(limits)[:2][np.argmax(np.abs(limits[:2]))]
                y = np.array(limits)[2:4][np.argmax(np.abs(limits[2:4]))]
                z = np.array(limits)[4:][np.argmax(np.abs(limits[4:]))]
                p.add_point_labels(
                    np.array([x, y, z]), [OCTANT_SIGNS[name]], text_color="black"
                )

        self._plotter = p

    @property
    def percent_buried_volume(self) -> float:
        """Deprecated attribute. Use 'fraction_buried_volume' instead."""
        warnings.warn(
            "'percent_buried_volume' is deprecated. Use 'fraction_buried_volume'.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.fraction_buried_volume

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


def cli(file: str) -> Any:
    """CLI for buried volume.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(BuriedVolume, elements, coordinates)



import setuptools

URL = "https://github.com/kjelljorner/morfeus"
DESCRIPTION = "A Python package for calculating molecular features"
LONG_DESCRIPTION = f"""\
{DESCRIPTION}. For more information, see the [project repository]({URL}).
"""

setuptools.setup(
    name="morfeus-ml",
    version="0.5.2",
    author="Kjell Jorner",
    author_email="kjell.jorner@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=URL,
    packages=["morfeus"],
    python_requires=">=3.8",
    install_requires=["fire", "numpy", "scipy"],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "morfeus=morfeus.__main__:main",
        ]
    },
)
